import multiprocessing as mp
import os
import shutil
from collections import namedtuple
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import peakutils
import pyfftw
import zarr
from matplotlib import pyplot as plt
from nptyping import Float32, NDArray, Shape
from tqdm import tqdm, trange

from .utils import get_slices, hsl2rgb, load_wisdom, save_wisdom


np1d = NDArray[Shape["*"], Float32]
np2d = NDArray[Shape["*,*"], Float32]
np3d = NDArray[Shape["*,*,*"], Float32]
np4d = NDArray[Shape["*,*,*,*"], Float32]
np5d = NDArray[Shape["*,*,*,*,*"], Float32]


class Pyzfn:
    """Postprocessing, visualization for amumax's zarr outputs"""

    def __init__(self, path: str) -> None:
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path Not Found : '{path}'")
        self.z: zarr.Group = zarr.open(path)
        self.path: Path = Path(path).absolute()
        self.name: str = self.path.name.replace(self.path.suffix, "")

    def __getitem__(self, item: str) -> Union[zarr.Array, zarr.Group]:
        return self.z[item]

    def __getattr__(self, name: str) -> Union[zarr.Array, zarr.Group, int, float, str]:
        if name in dir(self):
            return getattr(self, name)
        if name in dir(self.z):
            return getattr(self.z, name)
        if name in self.z.attrs:
            return self.z.attrs[name]
        raise KeyError(name)

    def __repr__(self) -> str:
        return f"Pyzfn('{self.name}')"

    def __str__(self) -> str:
        return f"Pyzfn('{self.name}')"

    def __name__(self) -> str:
        return self.name

    @property
    def pp(self) -> zarr.util.TreeViewer:
        """Pretty print the tree"""
        return self.z.tree(expand=True)

    @property
    def p(self) -> None:
        """Print the tree"""
        print(self.name, self.z.tree())

    def rm(self, dset: str) -> None:
        """
        Remove a group or dataset
        :param dset: str:

        """
        shutil.rmtree(f"{self.path}/{dset}", ignore_errors=True)

    def is_finished(self) -> bool:
        end_time: str = self.z.attrs["end_time"]
        return end_time != ""

    def is_running(self) -> bool:
        end_time: str = self.z.attrs["end_time"]
        return end_time == ""

    def mkdir(self, name: str) -> None:
        """
        Create nested directories
        :param name: str:

        """
        os.makedirs(f"{self.path}/{name}", exist_ok=True)

    def calc_disp(
        self, dset_in_str: str = "m", dset_out_str: str = "m", single_y: bool = False
    ) -> None:
        """
        Calculate the dispersion of `/dset_in_str` and saves it in `/disp/dset_out_str`
        :param dset_in_str: str:  (Default value = "m")
        :param dset_out_str: str:  (Default value = "m")

        """
        dset_in: zarr.Array = self[dset_in_str]
        dset_out: str = f"disp/{dset_out_str}"
        self.rm(dset_out)
        s = dset_in.shape
        arr_out = np.zeros((s[0] // 2 - 1, s[3] - 1, s[4]), dtype=np.float32)
        hann2d = np.sqrt(np.outer(np.hanning(s[0] - 1), np.hanning(s[3] - 1)))[
            :, :, None
        ]
        arr = pyfftw.empty_aligned((s[0] - 1, s[3] - 1, s[4]), dtype=np.complex64)
        load_wisdom(arr)
        fft = pyfftw.builders.fftn(
            arr,
            axes=(1, 0),
            threads=mp.cpu_count() // 2,
            planner_effort="FFTW_ESTIMATE",
            avoid_copy=True,
        )
        save_wisdom(arr)
        for iy in trange(
            dset_in.shape[2] // dset_in.chunks[2],
            leave=False,
            desc=f"Calculating the dispersion for `{self.name}`",
        ):
            arr[:] = dset_in[1:, 0, iy, 1:, :]
            arr *= hann2d
            out = fft()
            out -= np.average(out, axis=(0, 1))[None, None, :]
            out = out[: out.shape[0] // 2]
            out = np.fft.fftshift(out, axes=(1, 2))
            arr_out[:] += np.abs(out[: arr_out.shape[0]])
            if single_y:
                break
        self.z.create_dataset(f"{dset_out}/arr", data=arr_out)
        ts = dset_in.attrs["t"][:]
        freqs = np.fft.rfftfreq(len(ts), (ts[-1] - ts[0]) / len(ts))
        self.z.create_dataset(f"{dset_out}/freqs", data=freqs, chunks=None)
        kvecs = (
            np.fft.fftshift(np.fft.fftfreq(arr.shape[1], self.z.attrs["dx"]))
            * 2
            * np.pi
        )
        self.z.create_dataset(f"{dset_out}/kvecs", data=kvecs, chunks=None)

    def calc_modes(
        self,
        dset_in_str: str = "m",
        dset_out_str: str = "m",
        slices: Tuple[slice, slice, slice, slice, slice] = (
            slice(None),
            slice(None),
            slice(None),
            slice(None),
            slice(None),
        ),
    ) -> None:
        """
        Calculate the spin wave spectra and also the mode profiles of `/dset_in_str`
        and saves it in `/disp/dset_out_str`
        :param dset_in_str: str:  (Default value = "m")
        :param dset_out_str: str:  (Default value = "m")

        """
        dset_in: zarr.Array = self[dset_in_str]
        self.rm(f"modes/{dset_out_str}")
        self.rm(f"fft/{dset_out_str}")
        st, sz, sy, sx, sc = dset_in.shape
        _, cz, cy, cx, cc = dset_in.chunks
        ts = dset_in.attrs["t"][:]
        freqs = np.fft.rfftfreq(len(ts), (ts[-1] - ts[0]) / len(ts)) * 1e-9
        lenf = freqs.shape[0]
        self.z.create_dataset(f"fft/{dset_out_str}/freqs", data=freqs, chunks=False)
        dset_fft = self.z.create_dataset(
            f"fft/{dset_out_str}/spec", shape=(lenf, 3), chunks=False
        )
        self.z.create_dataset(f"modes/{dset_out_str}/freqs", data=freqs, chunks=False)
        dset_modes = self.z.create_dataset(
            f"modes/{dset_out_str}/arr",
            shape=(lenf, sz, sy, sx, sc),
            dtype=np.complex64,
            chunks=(50, cz, cy, cx, cc),
        )
        x0 = pyfftw.empty_aligned((st, cz, cy, cx, cc), dtype=np.complex64)
        load_wisdom(x0)
        fft = pyfftw.builders.fft(
            x0,
            axis=0,
            threads=mp.cpu_count() // 2,
            planner_effort="FFTW_ESTIMATE",
            avoid_copy=True,
        )
        save_wisdom(x0)
        fft_out = []
        zsls, ysls, xsls, csls = get_slices(dset_in.shape, dset_in.chunks, slices)
        chunk_nb = len(zsls) * len(ysls) * len(xsls) * len(csls)
        # import time

        # print(dset_in.chunks)
        with tqdm(total=chunk_nb, desc="Calculating SW modes", leave=False) as progress:
            for zsl in zsls:
                for ysl in ysls:
                    for xsl in xsls:
                        for csl in csls:
                            # print(f"{zsl=},{ysl=},{xsl=},{csl=}")
                            # t0 = time.time()
                            x0[:] = dset_in[slices[0], zsl, ysl, xsl, csl]
                            # t1 = time.time()
                            x0 -= np.average(x0, axis=0)[None, ...]
                            # t2 = time.time()
                            x0 *= np.hanning(x0.shape[0])[:, None, None, None, None]
                            # t3 = time.time()
                            x1 = fft()[:lenf]
                            # t4 = time.time()
                            dset_modes[slices[0], zsl, ysl, xsl, csl] = x1
                            # t5 = time.time()
                            x1 = np.abs(x1)
                            # t6 = time.time()
                            x1 = np.max(x1, axis=(1, 2, 3))
                            # t7 = time.time()
                            fft_out.append(x1)
                            # t8 = time.time()
                            # print(
                            #     f"loading slice:{t1-t0:.2f},average:{t2-t1:.2f},hanning:{t3-t2:.2f},fft:{t4-t3:.2f},saving modes:{t5-t4:.2f},abs:{t6-t5:.2f},max:{t7-t6:.2f},saving fft:{t8-t7:.2f}"
                            # )
                            progress.update()
        dset_fft[:] = np.max(np.array(fft_out), axis=0)

    def get_mode(self, dset: str, f: float, c: Union[int, None] = None) -> np4d:
        fi = int((np.abs(self[f"modes/{dset}/freqs"][:] - f)).argmin())
        arr: np4d = self[f"modes/{dset}/arr"][fi]
        if c is None:
            return arr
        else:
            return arr[..., c]

    def ispec(
        self,
        dset: str = "m",
        thres: float = 0.1,
        min_dist: int = 5,
        fmin: float = 0,
        fmax: float = 40,
        c: int = 0,
    ) -> None:
        Peak = namedtuple("Peak", ["idx", "freq", "amp"])

        def get_peaks(x: np1d, y: np1d) -> List[Peak]:
            idx = peakutils.indexes(y, thres=thres, min_dist=min_dist)
            peak_amp = [y[i] for i in idx]
            freqs = [x[i] for i in idx]
            return [Peak(i, f, a) for i, f, a in zip(idx, freqs, peak_amp)]

        def plot_spectra(ax: plt.Axes, x: np1d, y: np1d, peaks: List[Peak]) -> None:
            ax.plot(x, y)
            for _, freq, amp in peaks:
                ax.text(
                    freq,
                    amp + 0.03 * max(y),
                    f"{freq:.2f}",
                    # fontsize=5,
                    rotation=90,
                    ha="center",
                    va="bottom",
                )
            ax.set_title(self.name)
            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)

        def plot_modes(axes: plt.Axes, f: float) -> None:
            for ax in axes.flatten():
                ax.cla()
                ax.set(xticks=[], yticks=[])
            mode = self.get_mode("m", f)[0]
            extent = [
                0,
                mode.shape[1] * self.dx * 1e9,
                0,
                mode.shape[0] * self.dy * 1e9,
            ]
            for c in range(3):
                abs_arr = np.abs(mode[:, :, c])
                phase_arr = np.angle(mode[:, :, c])
                axes[0, c].imshow(
                    abs_arr,
                    cmap="inferno",
                    # vmin=0,
                    # vmax=mode_list_max,
                    norm=mpl.colors.LogNorm(vmin=5e-3),
                    extent=extent,
                    interpolation="None",
                    aspect="equal",
                )
                axes[1, c].imshow(
                    phase_arr,
                    aspect="equal",
                    cmap="hsv",
                    vmin=-np.pi,
                    vmax=np.pi,
                    interpolation="None",
                    extent=extent,
                )
                axes[2, c].imshow(
                    phase_arr,
                    aspect="equal",
                    alpha=abs_arr / abs_arr.max(),
                    cmap="hsv",
                    vmin=-np.pi,
                    vmax=np.pi,
                    interpolation="nearest",
                    extent=extent,
                )

        def get_spectrum() -> Tuple[np1d, np1d]:
            x = self[f"fft/{dset}/freqs"][:]
            y = self[f"fft/{dset}/spec"][:, c]
            x1 = np.abs(x - fmin).argmin()
            x2 = np.abs(x - fmax).argmin()
            return x[x1:x2], y[x1:x2]

        fig = plt.figure(constrained_layout=True, figsize=(15, 6))
        gs = fig.add_gridspec(1, 2)
        ax_spec = fig.add_subplot(gs[0, 0])
        x, y = get_spectrum()
        peaks = get_peaks(x, y)
        plot_spectra(ax_spec, x, y, peaks)
        axes_modes = gs[0, 1].subgridspec(3, 3).subplots()
        vline = ax_spec.axvline(10, ls="--", lw=0.8, c="#ffb86c")
        # plot_modes(axes_modes, peaks[0].freq)

        def onclick(event: mpl.backend_bases.Event) -> None:
            if event.inaxes == ax_spec:
                f = 10
                if event.button.name == "RIGHT":
                    freqs = [p.freq for p in peaks]
                    f = freqs[(np.abs(freqs - event.xdata)).argmin()]
                else:
                    f = event.xdata
                vline.set_data([f, f], [0, 1])
                plot_modes(axes_modes, f)
                fig.canvas.draw()

        fig.canvas.mpl_connect("button_press_event", onclick)

    def snapshot(
        self,
        dset: str = "m",
        z: int = 0,
        t: int = -1,
        ax: Optional[plt.Axes] = None,
        repeat: int = 1,
        zero: Optional[bool] = None,
    ) -> plt.Axes:
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(6, 6), dpi=100)
        else:
            fig = ax.figure
        arr = self[dset][t, z, :, :, :]
        if zero is not None:
            arr -= self[dset][zero, z, :, :, :]
        arr = np.tile(arr, (repeat, repeat, 1))
        arr = np.ma.masked_equal(arr, 0)
        u = arr[:, :, 0]
        v = arr[:, :, 1]
        z = arr[:, :, 2]

        alphas = -np.abs(z) + 1
        hsl = np.ones((u.shape[0], u.shape[1], 3), dtype=np.float32)
        hsl[:, :, 0] = np.angle(u + 1j * v) / np.pi / 2  # normalization
        hsl[:, :, 1] = np.sqrt(u**2 + v**2 + z**2)
        hsl[:, :, 2] = (z + 1) / 2
        rgb = hsl2rgb(hsl)
        stepx = max(int(u.shape[1] / 60), 1)
        stepy = max(int(u.shape[0] / 60), 1)
        scale = 1 / max(stepx, stepy)
        x, y = np.meshgrid(
            np.arange(0, u.shape[1], stepx) * float(self.dx) * 1e9,
            np.arange(0, u.shape[0], stepy) * float(self.dy) * 1e9,
        )
        antidots = np.ma.masked_not_equal(self[dset][0, 0, :, :, 2], 0)
        antidots = np.tile(antidots, (repeat, repeat))
        ax.quiver(
            x,
            y,
            u[::stepy, ::stepx],
            v[::stepy, ::stepx],
            alpha=alphas[::stepy, ::stepx],
            angles="xy",
            scale_units="xy",
            scale=scale,
        )
        ax.imshow(
            rgb,
            interpolation="None",
            origin="lower",
            cmap="hsv",
            vmin=-np.pi,
            vmax=np.pi,
            extent=[
                0,
                rgb.shape[1] * float(self.dx) * 1e9,
                0,
                rgb.shape[0] * float(self.dy) * 1e9,
            ],
        )
        ax.imshow(
            antidots,
            interpolation="None",
            origin="lower",
            cmap="Set1_r",
            extent=[
                0,
                arr.shape[1] * float(self.dx) * 1e9,
                0,
                arr.shape[0] * float(self.dy) * 1e9,
            ],
        )
        ax.set(title=self.name, xlabel="x (nm)", ylabel="y (nm)")
        fig.tight_layout()
        return ax

    def trim_modes(
        self, dset_in_str: str = "m", dset_out_str: str = "m", peak_xcut_min: int = 0
    ) -> None:
        self.rm(f"tmodes/{dset_in_str}")
        dset_in = self.z[f"modes/{dset_in_str}/arr"]
        all_peaks = []
        for c in range(3):
            spec = self.z.fft.m.spec[peak_xcut_min:, c]
            peaks = []
            for thres in np.linspace(0.1, 0.001):
                peaks = peakutils.indexes(spec / spec.max(), thres=thres, min_dist=2)
                if len(peaks) > 25:
                    break
            if len(peaks) == 0:
                print(f"No peaks found for {c=} in {self.name}")
            for p in peaks:
                all_peaks.append(p + peak_xcut_min)
        peaks = sorted(set(all_peaks))
        s = dset_in.shape
        self.z.create_dataset(
            f"tmodes/{dset_out_str}/freqs",
            data=np.array([self.z[f"modes/{dset_in_str}/freqs"][i] for i in peaks]),
            chunks=False,
            dtype=np.float32,
        )
        self.z.create_dataset(
            f"tmodes/{dset_out_str}/arr",
            data=np.array([dset_in[i] for i in peaks], np.complex64),
            chunks=(1, *s[1:]),
            dtype=np.complex64,
        )

    def ihist(self, dset: str = "m", xdata: str = "B_extz", ydata: str = "mz") -> None:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(15, 7))
        fig.subplots_adjust(bottom=0.16, top=0.94, right=0.99, left=0.08)
        b_ext = self.z[f"table/{xdata}"][:]
        m_avr = self.z[f"table/{ydata}"][:]
        passes = ["from 1T to -1T", "from -1T to 1T"]
        ax1.plot(b_ext[: len(b_ext) // 2], m_avr[: len(b_ext) // 2], label=passes[0])
        ax1.plot(
            b_ext[len(b_ext) // 2 :], m_avr[len(b_ext) // 2 :], label=passes[1], ls="--"
        )
        ax1.legend()
        B_sel = b_ext.shape[0] // 4
        vline = ax1.axvline(b_ext[B_sel], c="gray", ls=":")

        def onclick(event: mpl.backend_bases.Event) -> None:
            if event.inaxes == ax1:
                B_sel = np.abs(b_ext[: len(b_ext) // 2] - event.xdata).argmin()
                ax2.cla()
                if np.abs(m_avr[B_sel] - event.ydata) < np.abs(
                    m_avr[len(m_avr) // 2 :][::-1][B_sel] - event.ydata
                ):
                    self.snapshot(dset, t=B_sel, ax=ax2)
                    ax2.set_title(f"B_ext = {b_ext[B_sel]:.3f} T; from 1T to -1T")
                else:
                    self.snapshot(dset, t=len(b_ext) - B_sel, ax=ax2)
                    ax2.set_title(f"B_ext = {b_ext[B_sel]:.3f} T; from -1T to 1T")

                vline.set_data([b_ext[B_sel], b_ext[B_sel]], [0, 1])
                fig.canvas.draw()

        fig.canvas.mpl_connect("button_press_event", onclick)



import os
import pickle
import struct
from typing import List, Optional, Tuple, Dict

import matplotlib as mpl
import matplotlib.pyplot as plt
from ipympl.backend_nbagg import Canvas
import pyfftw
import numpy as np
import IPython
from nptyping import NDArray, Float, Shape, Int, Float32
import matplotx

m_type = NDArray[Shape["*,*,*,*,*"], Float32]


def wisdom_name_from_array(arr: m_type) -> str:
    shape = "_".join([str(i) for i in arr.shape])
    return os.path.expanduser(f"~/.cache/fftw/{shape}_{arr.dtype}")


def save_wisdom(arr: m_type) -> None:
    os.makedirs(os.path.expanduser("~/.cache/fftw/"), exist_ok=True)
    with open(wisdom_name_from_array(arr), "wb") as f:
        pickle.dump(pyfftw.export_wisdom(), f)


def load_wisdom(arr: m_type) -> bool:
    wisdom_path = wisdom_name_from_array(arr)
    if os.path.exists(wisdom_path):
        with open(wisdom_path, "rb") as f:
            pyfftw.import_wisdom(pickle.load(f))
        # print("Wisdom found.")
        return True
    else:
        # print("Wisdom not found, it might take a while optimizing FFTW.")
        return False


def make_cmap(
    min_color: Tuple[int, int, int, int],
    max_color: Tuple[int, int, int, int],
    mid_color: Optional[Tuple[int, int, int, int]] = None,
    transparent_zero: bool = False,
) -> mpl.colors.ListedColormap:
    cmap: NDArray[Shape["256, 4"], Float] = np.ones((256, 4))
    for i in range(4):
        if mid_color is None:
            cmap[:, i] = np.linspace(min_color[i], max_color[i], 256) / 256
        else:
            cmap[: 256 // 2, i] = (
                np.linspace(min_color[i], mid_color[i], 256 // 2) / 256
            )
            cmap[256 // 2 :, i] = (
                np.linspace(mid_color[i], max_color[i], 256 // 2) / 256
            )
    if transparent_zero:
        cmap[256 // 2, 3] = 0
    return mpl.colors.ListedColormap(cmap)


def save_ovf(
    path: str, arr: m_type, dx: float = 1e-9, dy: float = 1e-9, dz: float = 1e-9
) -> None:
    """Saves the given dataset for a given t to a valid OOMMF V2 ovf file"""

    def whd(s: str) -> None:
        s += "\n"
        f.write(s.encode("ASCII"))

    out = arr.astype("<f4").tobytes()

    xnodes, ynodes, znodes = arr.shape[2], arr.shape[1], arr.shape[0]
    xmin, ymin, zmin = 0, 0, 0
    xmax, ymax, zmax = xnodes * dx, ynodes * dy, znodes * dz
    xbase, ybase, _ = dx / 2, dy / 2, dz / 2
    valuedim = arr.shape[-1]
    valuelabels = "x y z"
    valueunits = "1 1 1"
    total_sim_time = "0"
    name = path.split("/")[-1]
    with open(path, "wb") as f:
        whd("# OOMMF OVF 2.0")
        whd("# Segment count: 1")
        whd("# Begin: Segment")
        whd("# Begin: Header")
        whd(f"# Title: {name}")
        whd("# meshtype: rectangular")
        whd("# meshunit: m")
        whd(f"# xmin: {xmin}")
        whd(f"# ymin: {ymin}")
        whd(f"# zmin: {zmin}")
        whd(f"# xmax: {xmax}")
        whd(f"# ymax: {ymax}")
        whd(f"# zmax: {zmax}")
        whd(f"# valuedim: {valuedim}")
        whd(f"# valuelabels: {valuelabels}")
        whd(f"# valueunits: {valueunits}")
        whd(f"# Desc: Total simulation time:  {total_sim_time}  s")
        whd(f"# xbase: {xbase}")
        whd(f"# ybase: {ybase}")
        whd(f"# zbase: {ybase}")
        whd(f"# xnodes: {xnodes}")
        whd(f"# ynodes: {ynodes}")
        whd(f"# znodes: {znodes}")
        whd(f"# xstepsize: {dx}")
        whd(f"# ystepsize: {dy}")
        whd(f"# zstepsize: {dz}")
        whd("# End: Header")
        whd("# Begin: Data Binary 4")
        f.write(struct.pack("<f", 1234567.0))
        f.write(out)
        whd("# End: Data Binary 4")
        whd("# End: Segment")


def load_ovf(path: str) -> NDArray[Shape["4"], Int]:
    with open(path, "rb") as f:
        dims = np.array([0, 0, 0, 0])
        while True:
            line = f.readline().strip().decode("ASCII")
            if "valuedim" in line:
                dims[3] = int(line.split(" ")[-1])
            if "xnodes" in line:
                dims[2] = int(line.split(" ")[-1])
            if "ynodes" in line:
                dims[1] = int(line.split(" ")[-1])
            if "znodes" in line:
                dims[0] = int(line.split(" ")[-1])
            if "Begin: Data" in line:
                break
        count = int(dims[0] * dims[1] * dims[2] * dims[3] + 1)
        arr = np.fromfile(f, "<f4", count=count)[1:].reshape(dims)
    return arr


def get_ovf_parms(path: str) -> Dict[str, float]:
    with open(path, "rb") as f:
        parms: Dict[str, float] = {}
        while True:
            line = f.readline().strip().decode("ASCII")
            if "valuedim" in line:
                parms["comp"] = int(line.split(" ")[-1], 10)
            if "xnodes" in line:
                parms["Nx"] = int(line.split(" ")[-1])
            if "ynodes" in line:
                parms["Ny"] = int(line.split(" ")[-1])
            if "znodes" in line:
                parms["Nz"] = int(line.split(" ")[-1])
            if "xstepsize" in line:
                parms["dx"] = float(line.split(" ")[-1])
            if "ystepsize" in line:
                parms["dy"] = float(line.split(" ")[-1])
            if "zstepsize" in line:
                parms["dz"] = float(line.split(" ")[-1])
            if "Begin: Data" in line:
                break
    return parms


def get_slices(
    shape: Tuple[int, int, int, int, int],
    chunks: Tuple[int, int, int, int, int],
    slices: Tuple[slice, slice, slice, slice, slice],
) -> List[List[slice]]:
    out: List[List[slice]] = [[], [], [], []]
    for i, (s, c, sl) in enumerate(list(zip(shape, chunks, slices))[1:]):
        tmp_list: List[List[int]] = []
        for pt in list(range(s))[sl]:
            chunk_nb = pt // c
            if chunk_nb >= len(tmp_list):
                tmp_list.append([])
            tmp_list[chunk_nb].append(pt)
        for sublist in tmp_list:
            out[i].append(slice(min(sublist), max(sublist) + 1, sl.step))
    return out


def load_mpl_style(skip_style: bool = False) -> None:
    ipy = IPython.get_ipython()  # type: ignore
    if ipy is not None:
        ipy.run_cell_magic(
            "html",
            "",
            "<style> .cell-output-ipywidget-background {background-color: transparent !important;}</style>",
        )
        ipy.run_line_magic("matplotlib", "widget")
        ipy.run_line_magic("load_ext", "autoreload")
        ipy.run_line_magic("autoreload", "2")

        plt.rcParams["figure.max_open_warning"] = 1000

        Canvas.header_visible.default_value = False
        Canvas.footer_visible.default_value = True
    if not skip_style:
        plt.style.use(matplotx.styles.dracula)
        plt.rcParams["figure.figsize"] = [14, 5]
        plt.rcParams["figure.autolayout"] = True
        # plt.rcParams["axes.grid"] = True
        plt.rcParams["axes.grid.axis"] = "both"


def save_current_mplstyle(path: str) -> None:
    from matplotlib import pyplot as plt

    newlines = []
    for k, v in plt.rcParams.items():
        if k in [
            "backend",
            "backend_fallback",
            "date.epoch",
            "docstring.hardcopy",
            "figure.max_open_warning",
            "figure.raise_window",
            "interactive",
            "savefig.directory",
            "timezone",
            "tk.window_focus",
            "toolbar",
            "webagg.address",
            "webagg.open_in_browser",
            "webagg.port",
            "webagg.port_retries",
        ]:
            continue
        if isinstance(v, list):
            v = [str(i) for i in v]
            v = ", ".join(v)
        v = str(v)
        if len(v) > 4:
            if v[0] == "#":
                v = v[1:]
        if k == "axes.prop_cycle":
            v = v.replace("#", "")
        if k == "grid.color":
            v = '"' + v + '"'
        if k == "lines.dash_joinstyle":
            v = "round"
        if k == "lines.dash_capstyle":
            v = "butt"
        if k == "lines.solid_capstyle":
            v = "projecting"
        if k == "lines.solid_joinstyle":
            v = "round"
        if k == "savefig.bbox":
            v = "tight"
        newlines.append(k + ": " + v + "\n")

    with open(path, "w") as f:
        f.writelines(newlines)


def hsl2rgb(hsl: NDArray[Shape["*,*,3"], Float32]) -> NDArray[Shape["*,*,3"], Float32]:
    h = hsl[..., 0] * 360
    s = hsl[..., 1]
    l = hsl[..., 2]

    rgb = np.zeros_like(hsl)
    for i, n in enumerate([0, 8, 4]):
        k = (n + h / 30) % 12
        a = s * np.minimum(l, 1 - l)
        k = np.minimum(k - 3, 9 - k)
        k = np.clip(k, -1, 1)
        rgb[..., i] = l - a * k
    rgb = np.clip(rgb, 0, 1)
    return rgb
